{{
  Time Format Object

  These functions operate on two longs formatted as follows:

  Time = sec (0-59) + min (0-59) << 8 + hour (0-23) << 16  
  Date = day (1-7) + date (1-31) << 8 + month (1-12) << 16 + year (0-99) << 24

  Note that these octets (binary) correspond to the byte data (BCD) returned
  from a DS1307 RTC chip.

  When setting the time, the set field index corresponds to the DS1307 register
  in which the value is found:

  0=sec 1=min 2=hour 3=day of week 4=date 5=month 6=year-2000

  This object follows the methodology of the strfmt object and so can be used
  by multiple objects and cog threads without penalty.
}}

con

  'color bytes to set and clear emphasis on set format
  '
  emset = %1111
  emclr = %1000

  'style bits
  '
  st_hr_12       = %0000_0001
  st_show_sec    = %0000_0010
  st_show_day    = %0000_0100
  st_spell_month = %0000_1000
  st_date_ddmm   = %0001_0000
  st_year_2d     = %0010_0000

dat
{
  This is a dat instead of var so that only one instance will
  be used no matter how many places this object is used
}
style   byte  st_show_sec + st_show_day + st_spell_month
  
obj
  sf        : "strfmt"
  bsf       : "bufstrfmt"

pub getstyle

  return style

pub setstyle(value)

  style := value

{
  In order to avoid churning the bufstrfmt return buffer these
  functions manually request a return buffer large enough to
  accommodate the largest possible result and then build the
  result by appending in place.  This does introduce the issue
  of how long the BufStrFmt buffers can be trusted not to be
  overwritten.  In this application we have a very large buffer
  thanks to the keyboard PASM image so this is quite safe.  
}
pub timefmt(atime) : ptr | h
{
  Return a formatted string for the system standard atime long
}
  ptr := bsf.bufalloc(15)
  h := hour(atime)
  if style & st_hr_12
    if h > 12
      h -= 12
    elseif h == 0
      h := 12
  sf.cat(ptr, fmt2(h))
  sf.chcat(ptr, ":")
  sf.cat(ptr, fmt2(minute(atime)))
  if style & st_show_sec
    sf.chcat(ptr, ":")
    sf.cat(ptr, fmt2(second(atime)))
  if style & st_hr_12
    if hour(atime) > 12
      sf.cat(ptr, string(" PM"))
    else
      sf.cat(ptr, string(" AM"))

pub datefmt(adate) : ptr
{
  Return a formatted string for the system standard adate long
}
  ptr := bsf.bufalloc(20)
  if style & st_show_day
    sf.cat(ptr, dayname(day(adate)))
    sf.chcat(ptr, " ")
  if style & st_date_ddmm
    sf.cat(ptr, datedatefmt(adate))
    sf.chcat(ptr, datesep1fmt)
    sf.cat(ptr, datemonthfmt(adate))
  else
    sf.cat(ptr, datemonthfmt(adate))
    sf.chcat(ptr, datesep1fmt)
    sf.cat(ptr, datedatefmt(adate))
  sf.cat(ptr, datesep2fmt)
  if not style & st_year_2d
    sf.cat(ptr, string("20"))
  sf.cat(ptr, fmt2(year(adate)))

pri datemonthfmt(adate)

  if style & st_spell_month
    result := monthname(month(adate))
  else
    result := fmt2(month(adate))
    
pri datedatefmt(adate)

  result := fmt2(date(adate))

pri datesep1fmt
' separator between date and month as character code
  if style & st_spell_month
    return " "
  else
    return "-"

pri datesep2fmt : ptr
' separator before year as string
  if style & st_spell_month
    if style & constant(st_date_ddmm | st_year_2d)
      ptr := bsf.ch(" ")
    else 
      ptr := bsf.str(string(", "))
  else
    ptr := bsf.ch("-")

{
  When setting the clock, the seconds and day are always displayed
  regardless of style settings.  aset conditionally wraps one of
  the fields in emphasis codes.
}
pub setclkfmt(atime, adate, aset) : ptr | h

  ptr := bsf.bufalloc(40)
  sf.cat(ptr, string($C, emclr))
  h := hour(atime)
  if style & st_hr_12
    if h > 12
      h -= 12
    elseif h == 0
      h := 12
  ifset(ptr, fmt2(h), aset == 2)

  sf.chcat(ptr, ":")
  ifset(ptr, fmt2(minute(atime)), aset == 1)
  sf.chcat(ptr, ":")
  ifset(ptr, fmt2(second(atime)), aset == 0)
  if style & st_hr_12
    sf.chcat(ptr, " ")
    if hour(atime) > 12
      ifset(ptr, string("PM"), aset == 2)
    else
      ifset(ptr, string("AM"), aset == 2)

  sf.chcat(ptr, " ")
  
  ifset(ptr, dayname(day(adate)),aset == 3)
  sf.chcat(ptr, " ")
  if style & st_date_ddmm
    
    ifset(ptr, datedatefmt(adate), aset == 4)
    sf.chcat(ptr, datesep1fmt)
    ifset(ptr, datemonthfmt(adate), aset == 5)
  else
    ifset(ptr, datemonthfmt(adate), aset == 5)
    sf.chcat(ptr, datesep1fmt)
    ifset(ptr, datedatefmt(adate), aset == 4)
  sf.cat(ptr, datesep2fmt)
  ifset(ptr, bsf.strcat(string("20"),fmt2(year(adate))), aset == 6)

pri ifset(ptr, addstr, doset)
{
  sf.strcat the addstr string, wrapped in emphasis
  codes if doset is true
}  '
  if doset
    sf.cat(ptr, string($C, emset))
  sf.cat(ptr, addstr)
  if doset
    sf.cat(ptr, string($C, emclr))

pub dayname(aday)
  '
  return bsf.lstr(string("???SunMonTueWedThuFriSat") + aday * 3, 3)

pub monthname(amonth)
  '
  return bsf.lstr(string("???JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC") + amonth * 3, 3) 

pri fmt2(value)
  '
  return bsf.fdec(value, 2, 2)
{
  Utilities to extract and set the individual octets of the formatted time and date longs
}
pub second(atime)
  '
  return atime & $FF

pub minute(atime)
  '
  return (atime >> 8) & $FF
  
pub hour(atime)
  '
  return (atime >> 16) & $FF

pub day(adate)
  '
  return adate & $FF

pub date(adate)
  '
  return (adate >> 8) & $FF

pub month(adate)
  '
  return (adate >> 16) & $FF

pub year(adate)
  '
  return (adate >> 24) & $FF

pub setsecond(atime, value)
  '
  return (atime & !$FF) | (value & $FF)

pub setminute(atime, value)
  '
  return (atime & !$FF00) | ((value & $FF) << 8)
  
pub sethour(atime, value)
  '
  return (atime & !$FF0000) | ((value & $FF) << 16)

pub setday(adate, value)
  '
  return (adate & !$FF) | (value & $FF)

pub setdate(adate, value)
  '
  return (adate & !$FF00) | ((value & $FF) << 8)

pub setmonth(adate, value)
  '
  return (adate & !$FF0000) | ((value & $FF) << 16)

pub setyear(adate, value)
  '
  return (adate & !$FF000000) | ((value & $FF) << 24)

   